import os
from pathlib import Path

import numpy as np

import czifile
import tifffile

class GenericImageFileAccessor(object):
    def __init__(self, fpath: Path):
        if not os.path.exists(fpath):
            raise FileAccessorError(f'Could not find file at {fpath}')

class CziImageFileAccessor(GenericImageFileAccessor):
    def __init__(self, fpath: Path):
        super().__init__(fpath)
        self.fpath = fpath
        try:
            cf = czifile.CziFile(fpath)
            self.czifile = cf
        except:
            raise FileAccessorError(f'Unable to access data in {fpath}')

        sd = {ch: cf.shape[cf.axes.index(ch)] for ch in cf.axes}
        if sd['S'] > 1 or sd['T'] > 1:
            raise FileShapeError(f'Cannot handle image with multiple positions or time points: {sd}')
        self.is_3d = (sd['Z'] > 1)
        self.shape_dict = sd
        self.chroma = sd['C']

        self.data = np.moveaxis(
            cf.asarray(),
            [cf.axes.index(ch) for ch in ['X', 'Y', 'C', 'Z']],
            [0, 1, 2, 3]
        )

    def __del__(self):
        self.czifile.close()

class WriteableTiffFileAccessor(GenericImageFileAccessor):
    def __init__(self, fpath: Path):
        self.fpath = fpath

    def write(self, data):
        try:
            tifffile.imwrite(self.fpath, data)
        except:
            raise FileWriteError(f'Unable to write data to file')


def generate_file_accessor(fpath):
    if str(fpath).upper().endswith('.CZI'):
        return CziImageFileAccessor(fpath)
    else:
        raise FileAccessorError(f'Could not match a file accessor with {fpath}')

def Error(Exception):
    pass

def FileAccessorError(Error):
    pass

def FileNotFoundError(Error):
    pass

def FileShapeError(Error):
    pass

def FileWriteError(Error):
    pass