import json
import os

from pathlib import Path
from time import strftime, localtime

from conf.server import paths
from model_server.model import Model
from model_server.share import SharedImageDirectory
from model_server.workflow import WorkflowRunRecord

def create_manifest_json():
    pass

class Session(object):
    """
    Singleton class for a server session that persists data between API calls
    """
    inbound = SharedImageDirectory(paths['images']['inbound'])
    outbound = SharedImageDirectory(paths['images']['outbound'])
    where_records = Path(paths['logs']['session'])

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Session, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        print('Initializing session')
        self.models = {} # model_id : model object
        self.manifest = [] # paths to data as well as other metadata from each inference run
        self.session_id = self.create_session_id(self.where_records)
        self.session_log = self.where_records / f'{self.session_id}.log'
        self.log_event('Initialized session')
        self.manifest_json = self.where_records / f'{self.session_id}-manifest.json'
        open(self.manifest_json, 'w').close() # instantiate empty json file

    @staticmethod
    def create_session_id(look_where: Path) -> str:
        """
        Autogenerate a session ID by incrementing from a list of log files.
        """
        yyyymmdd = strftime('%Y%m%d', localtime())
        idx = 0
        while os.path.exists(look_where / f'{yyyymmdd}-{idx:04d}.log'):
            idx += 1
        return f'{yyyymmdd}-{idx:04d}'

    def log_event(self, event: str):
        """
        Write an event string this session's log file.
        """
        timestamp = strftime('%m/%d/%Y, %H:%M:%S', localtime())
        with open(self.session_log, 'w+') as fh:
            fh.write(f'{timestamp} -- {event}')

    def record_workflow_run(self, record: WorkflowRunRecord or None):
        """
        Append a JSON describing inference data to this session's manifest
        """
        with open(self.manifest_json, 'w+') as fh:
            json.dump(record.dict(), fh)

    def load_model(self, model_id: str) -> bool:
        """
        Load an instance of first model class that matches model_id string
        :param model_id:
        :return: True if model successfully loaded, False if not
        """
        for mc in Model.__subclasses__():
            if hasattr(mc, 'model_id') and getattr(mc, 'model_id') == model_id:
                mi = mc()
                assert mi.loaded
                self.models['model_id'] = mi
                return True
        return False

    def describe_models(self) -> dict:
        return {
            k: self.models[k].__class__.__name__
            for k in self.models.keys()
        }

    def restart(self):
        self.__init__()

class Error(Exception):
    pass

class InferenceRecordError(Error):
    pass